from dataclasses import dataclass
from typing import List
import uuid


@dataclass
class TileLayerComponent:
    def __init__(
        self,
        url: str = '',
        name: str = '',
        description: str = '',
        visible: bool = True,
        opacity: float = 1.0,
    ):
        self.url = url
        self.name = name
        self.description = description
        self.visible = visible
        self.opacity = opacity

    def convert_to_dataclass(self):
        id = uuid.uuid4().hex
        return TileLayer(id=id, url=self.url, name=self.name, description=self.description, visible=self.visible, opacity=self.opacity)


@dataclass()
class TileLayer:
    id: str
    url: str
    name: str
    description: str
    visible: bool
    opacity: float



from typing import Union

from .bar_chart import BarChart
from .bar_chart import BarChartComponentInfo
from .draw_feature import DrawFeature
from .draw_feature import DrawFeatureComponentInfo
from .line_chart import LineChart
from .line_chart import LineChartComponentInfo
from .multiselect import Multiselect
from .multiselect import MultiselectComponentInfo
from .number import Number
from .number import NumberComponentInfo
from .text import Text
from .text import TextComponentInfo
from .select import Select
from .select import SelectComponentInfo
from .display import Display
from .display import DisplayComponentInfo

# TODO add interface to inputs
GreppoCharts = Union[BarChart, LineChart]
GreppoChartNames = [i.proxy_name() for i in GreppoCharts.__args__]

GreppoInputs = Union[Display, Number, Text, Select, Multiselect, DrawFeature]
GreppoInputsNames = [i.proxy_name() for i in GreppoInputs.__args__]

ComponentInfo = Union[
    DisplayComponentInfo,
    NumberComponentInfo,
    TextComponentInfo,
    SelectComponentInfo,
    MultiselectComponentInfo,
    DrawFeatureComponentInfo,
    BarChartComponentInfo,
    LineChartComponentInfo,
]



import logging

import click

from .greppo import GreppoApp
from .greppo_server import GreppoServer


@click.command()
@click.argument("command")
@click.argument("path_to_script")
@click.argument("host", default="0.0.0.0")
@click.argument("port", default="8080")
def wrap_and_run_script(command, path_to_script, host, port):
    """Run a Greppo COMMAND with PATH_TO_SCRIPT_NAME

    PATH_TO_SCRIPT is the path to the script
    COMMANDS is one of [serve]
    """
    if command == "serve":
        gpo = GreppoApp()
        server = GreppoServer(gr_app=gpo, user_script=path_to_script)

        server.run(host=host, port=int(port))
    else:
        logging.error("Command {} not supported".format(command))


if __name__ == "__main__":
    wrap_and_run_script()



from dataclasses import dataclass
from typing import List, Union, Dict
from .tile_layer import TileLayer
import ee
import uuid


@dataclass
class EarthEngineLayerComponent:
    def __init__(
        self,
        ee_object: Union[ee.Image, ee.ImageCollection, ee.FeatureCollection, ee.Feature, ee.Geometry],
        name: str = '',
        description: str = '',
        visible: bool = True,
        vis_params: Dict = {},
        opacity: float = 1.0,
    ):
        self.name = name
        self.description = description
        self.visible = visible
        self.opacity = opacity

        if (
            isinstance(ee_object, ee.geometry.Geometry)
            or isinstance(ee_object, ee.feature.Feature)
            or isinstance(ee_object, ee.featurecollection.FeatureCollection)
        ):
            features = ee.FeatureCollection(ee_object)

            width = vis_params.get('width', 2)
            color = vis_params.get('color', '000000')

            image_fill = features.style(**{"fillColor": color}).updateMask(
                ee.Image.constant(0.5)
            )
            image_outline = features.style(
                **{"color": color, "fillColor": "00000000", "width": width}
            )

            self.ee_object_image = image_fill.blend(image_outline)
        elif isinstance(ee_object, ee.image.Image):
            self.ee_object_image = ee_object
        elif isinstance(ee_object, ee.imagecollection.ImageCollection):
            self.ee_object_image = ee_object.mosaic()

        self.vis_params = vis_params

    def convert_to_dataclass(self):
        id = uuid.uuid4().hex
        map_id_dict = ee.Image(self.ee_object_image).getMapId(self.vis_params)
        url = map_id_dict["tile_fetcher"].url_format
        return TileLayer(id=id, url=url, name=self.name, description=self.description, visible=self.visible, opacity=self.opacity)



from dataclasses import dataclass
from typing import List, Union
import uuid
from click import style
from geopandas import GeoDataFrame as gdf
from numpy import arange
from ..colorbrewer import get_palette


@dataclass
class VectorLayerComponent:
    def __init__(
        self,
        data: gdf,
        name: str,
        description: str = '',
        style: dict = {},
        visible: bool = True
    ):
        self.data = data
        self.name = name
        self.description = description
        self.style = style
        self.visible = visible

        if 'choropleth' in style:
            choropleth_style = style['choropleth']
            if 'key_on' not in choropleth_style:
                raise ValueError(
                    '"key_on" not specified in style: "choropleth" for vector_layer')
            else:
                key_on = choropleth_style['key_on']
            if key_on in data:
                choropleth_on_data = data[key_on]
            else:
                raise ValueError(
                    f'"{key_on}" is an invalid key for the passed in DataFrame')
            bin_min = choropleth_on_data.min()
            bin_max = choropleth_on_data.max()
            if 'bins' in choropleth_style:
                bins = choropleth_style['bins']
                if isinstance(bins, int):
                    if bins <= 0:
                        ValueError(
                            f'"{bins}" is a invalid for "bins" in style: "choropleth", "bins" must be >= 1')
                    bin_step = (bin_min+bin_max)/bins
                    bins = list(arange(bin_min, bin_max, bin_step))
                    choropleth_style['bins'] = bins
                elif isinstance(bins, List):
                    if not all(bin_min <= x <= bin_max for x in bins[1:]):
                        raise ValueError(
                            'Values in "bins" do not match the values in the DataFrame')
                    if not all(x <= y for x, y in zip(bins[:-1], bins[1:])):
                        raise ValueError('Values in "bins" are not increasing')
            else:
                bin_step = (bin_min+bin_max)/6
                bins = list(arange(bin_min, bin_max, bin_step))
                choropleth_style['bins'] = bins
            bins_length = len(bins)
            if 'palette' in choropleth_style:
                if isinstance(choropleth_style['palette'], str):
                    choropleth_style['palette'] = get_palette(
                        bins_length, choropleth_style['palette'])
                elif isinstance(choropleth_style['palette'], List[str]):
                    if bins_length == len(choropleth_style['palette']):
                        pass
                        # TODO Check if each item is a valid color Hex or RGB code.
                    else:
                        raise ValueError(
                            '"bins" length and "palette" length must match')
            else:
                choropleth_style['palette'] = get_palette(bins_length, 'Purples')

            choropleth_style['bins'] = list(reversed(choropleth_style['bins'])) 
            choropleth_style['palette'] = list(reversed(choropleth_style['palette'])) 
            self.style['choropleth'] = choropleth_style

    def convert_to_dataclass(self):
        id = uuid.uuid4().hex
        gdf_bounds = self.data.total_bounds
        bounds = [[gdf_bounds[1], gdf_bounds[0]],
                  [gdf_bounds[3], gdf_bounds[2]]]
        return VectorLayer(id=id, data=self.data, name=self.name, description=self.description, style=self.style, visible=self.visible, bounds=bounds)


@dataclass
class VectorLayer:
    id: str
    data: gdf
    name: str
    description: str
    style: dict
    visible: bool
    bounds: List[List]



from dataclasses import dataclass
from typing import List, Union
import xyzservices
import uuid


@dataclass
class BaseLayerComponent:
    def __init__(
        self,
        provider: Union[str, xyzservices.TileProvider] = '',
        name: str = '',
        visible: bool = False,
        url: str = '',
        attribution: str = '',
        subdomains: Union[str, List[str]] = '',
        min_zoom: int = 0,
        max_zoom: int = 18,
        bounds: List[List[int]] = [],
    ):
        self.provider = provider
        self.name = name
        self.visible = visible
        self.url = url
        self.attribution = attribution
        self.subdomains = subdomains
        self.min_zoom = min_zoom
        self.max_zoom = max_zoom
        self.bounds = bounds

    def convert_to_dataclass(self):
        id = uuid.uuid4().hex
        tile_provider = None

        if self.provider and isinstance(self.provider, str):
            try:
                tile_provider = xyzservices.providers.query_name(self.provider)
            except ValueError:
                raise ValueError(
                    f"No matching provider found for: '{self.provider}'.")

        if isinstance(self.provider, xyzservices.TileProvider):
            tile_provider = self.provider

        if tile_provider is not None:
            name = self.name if self.name else tile_provider.name.replace(
                ".", " - ")
            url = self.url if self.url else tile_provider.build_url(
                fill_subdomain=False)
            attribution = self.attribution if self.attribution else tile_provider.attribution if 'attribution' in tile_provider else tile_provider.html_attribution
            subdomains = self.subdomains if self.subdomains else tile_provider.subdomains if 'subdomains' in tile_provider else ''

            return BaseLayer(id, name=name, visible=self.visible, url=url, subdomains=subdomains, attribution=attribution)

        else:
            if not self.name:
                raise ValueError(
                    "If 'provider' is not passed for 'base_layer', please pass in 'name'.")
            if not self.url:
                raise ValueError(
                    "If 'provider' is not passed for 'base_layer', please pass in 'url'.")

            return BaseLayer(id, name=self.name, visible=self.visible, url=self.url, subdomains=self.subdomains, attribution=self.attribution)


@dataclass()
class BaseLayer:
    id: str
    name: str
    visible: bool
    url: str
    subdomains: Union[str, List[str]]
    attribution: str



"""OpenStreetMap utilities Python.

Modified version of github.com/rossant/smopy .
"""
# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import logging

import numpy as np

# -----------------------------------------------------------------------------
# OSM functions
# -----------------------------------------------------------------------------
def correct_box(box, z):
    """Get good box limits"""
    x0, y0, x1, y1 = box
    new_x0 = max(0, min(x0, x1))
    new_x1 = min(2 ** z - 1, max(x0, x1))
    new_y0 = max(0, min(y0, y1))
    new_y1 = min(2 ** z - 1, max(y0, y1))

    return (new_x0, new_y0, new_x1, new_y1)


def get_box_size(box):
    """Get box size"""
    x0, y0, x1, y1 = box
    sx = abs(x1 - x0) + 1
    sy = abs(y1 - y0) + 1
    return (sx, sy)


# -----------------------------------------------------------------------------
# Functions related to coordinates
# -----------------------------------------------------------------------------
def deg2num(latitude, longitude, zoom, do_round=True):
    """Convert from latitude and longitude to tile numbers.

    If do_round is True, return integers. Otherwise, return floating point
    values.

    Source: http://wiki.openstreetmap.org/wiki/Slippy_map_tilenames#Python

    """
    lat_rad = np.radians(latitude)
    n = 2.0 ** zoom
    if do_round:
        f = np.floor
    else:
        f = lambda x: x
    xtile = f((longitude + 180.0) / 360.0 * n)
    ytile = f((1.0 - np.log(np.tan(lat_rad) + (1 / np.cos(lat_rad))) / np.pi) / 2.0 * n)
    if do_round:
        if isinstance(xtile, np.ndarray):
            xtile = xtile.astype(np.int32)
        else:
            xtile = int(xtile)
        if isinstance(ytile, np.ndarray):
            ytile = ytile.astype(np.int32)
        else:
            ytile = int(ytile)
    return (xtile, ytile)


def get_tile_box(box_latlon, z):
    """Convert a box in geographical coordinates to a box in
    tile coordinates (integers), at a given zoom level.

    box_latlon is lat0, lon0, lat1, lon1.

    """
    lat0, lon0, lat1, lon1 = box_latlon
    x0, y0 = deg2num(lat0, lon0, z)
    x1, y1 = deg2num(lat1, lon1, z)
    return (x0, y0, x1, y1)


def _box(*args):
    """Return a tuple (lat0, lon0, lat1, lon1) from a coordinate box that
    can be specified in multiple ways:

    A. box((lat0, lon0))  # nargs = 1
    B. box((lat0, lon0, lat1, lon1))  # nargs = 1
    C. box(lat0, lon0)  # nargs = 2
    D. box((lat0, lon0), (lat1, lon1))  # nargs = 2
    E. box(lat0, lon0, lat1, lon1)  # nargs = 4

    """
    nargs = len(args)
    assert nargs in (1, 2, 4)
    pos1 = None

    # Case A.
    if nargs == 1:
        assert hasattr(args[0], "__len__")
        pos = args[0]
        assert len(pos) in (2, 4)
        if len(pos) == 2:
            pos0 = pos
        elif len(pos) == 4:
            pos0 = pos[:2]
            pos1 = pos[2:]

    elif nargs == 2:
        # Case C.
        if not hasattr(args[0], "__len__"):
            pos0 = args[0], args[1]
        # Case D.
        else:
            pos0, pos1 = args[0], args[1]

    # Case E.
    elif nargs == 4:
        pos0 = args[0], args[1]
        pos1 = args[2], args[3]

    if pos1 is None:
        pos1 = pos0

    return (pos0[0], pos0[1], pos1[0], pos1[1])


def extend_box(box_latlon, margin=0.1):
    """Extend a box in geographical coordinates with a relative margin."""
    (lat0, lon0, lat1, lon1) = box_latlon
    lat0, lat1 = min(lat0, lat1), max(lat0, lat1)
    lon0, lon1 = min(lon0, lon1), max(lon0, lon1)
    dlat = max((lat1 - lat0) * margin, 0.0005)
    dlon = max((lon1 - lon0) * margin, 0.0005 / np.cos(np.radians(lat0)))
    return (
        max(lat0 - dlat, -80),
        max(lon0 - dlon, -180),
        min(lat1 + dlat, 80),
        min(lon1 + dlon, 180),
    )


# -----------------------------------------------------------------------------
# Main Map class
# -----------------------------------------------------------------------------
class Map(object):
    def __init__(self, *args, **kwargs):
        """Create and fetch the map with a given box in geographical
        coordinates.

        """
        z = kwargs.get("z", 18)
        margin = kwargs.get("margin", 0.05)

        self.tilesize = kwargs.get("tilesize", 256)
        self.maxtiles = kwargs.get("maxtiles", 16)
        self.verbose = kwargs.get("verbose", False)

        box = _box(*args)
        if margin is not None:
            box = extend_box(box, margin)
        self.box = box

        self.z = self.get_allowed_zoom(z)
        if z > self.z:
            if self.verbose:
                logging.info(
                    "Lowered zoom level to keep map size reasonable. "
                    "(z = %d)" % self.z
                )
        else:
            self.z = z

    def get_allowed_zoom(self, z=18):
        box_tile = get_tile_box(self.box, z)
        box = correct_box(box_tile, z)
        sx, sy = get_box_size(box)
        if sx * sy >= self.maxtiles:
            z = self.get_allowed_zoom(z - 1)
        return z



from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import List
from typing import Optional


@dataclass
class Dataset:
    label: str
    backgroundColor: str
    data: List[Any]


@dataclass
class ChartData:
    labels: List[Any]
    datasets: List[Dataset]


# TODO doesn't handle multiple datasets.
@dataclass
class LineChart:
    def __init__(
        self,
        name: str,
        description: Optional[str],
        x: List,
        y: List,
        color: str = "#000000",
        input_updates: Dict[str, Any] = {},
    ):
        self.input_name = name
        self.description = description
        self.input_updates = input_updates
        
        _, label = name.split("_")
        dataset = Dataset(label=label, data=y, backgroundColor=color)
        self.chartdata = ChartData(labels=x, datasets=[dataset])

    def get_value(self):
        return None

    def convert_to_component_info(self):
        _id, name = self.input_name.split("_")
        _type = LineChart.__name__

        return LineChartComponentInfo(
            name=name,
            id=_id,
            type=_type,
            description=self.description,
            chartdata=self.chartdata,
        )

    @classmethod
    def proxy_name(cls):
        return "line_chart"

    def __repr__(self):
        return str(self.get_value())

    def __str__(self):
        return self.__repr__()


@dataclass
class LineChartComponentInfo:
    name: str
    id: str
    type: str
    description: str
    chartdata: ChartData



from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import List
from typing import Union


SELECT_TYPES = Union[int, float, str, bool]


@dataclass
class Select:
    def __init__(
        self,
        name: str,
        options: List[SELECT_TYPES],
        default: SELECT_TYPES,
        input_updates: Dict[str, Any] = {},
    ):
        self.input_name = name
        self.options = options
        self.default = default

        self.input_updates = input_updates

    def get_value(self):
        id, name = self.input_name.split("_")
        return self.input_updates.get(name, self.default)

    def convert_to_component_info(self):
        _id, name = self.input_name.split("_")
        _type = Select.__name__
        value = str(self.get_value())

        return SelectComponentInfo(
            id=_id, name=name, type=_type, value=value, options=self.options
        )

    @classmethod
    def proxy_name(cls):
        return "select"

    def __repr__(self):
        return str(self.get_value())

    def __str__(self):
        return self.__repr__()


@dataclass
class SelectComponentInfo:
    id: str
    name: str
    type: str
    value: Union[
        int, float, str, bool
    ]  # This can represent the default user supplied value
    options: List[Any]



import logging
import os
from distutils.sysconfig import get_python_lib
from functools import partial

import uvicorn
from greppo import GreppoApp
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.cors import CORSMiddleware
from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette.responses import Response
from starlette.routing import Mount
from starlette.routing import Route
from starlette.routing import WebSocketRoute
from starlette.staticfiles import StaticFiles

from .user_script_utils import script_task


async def api_endpoint(user_script: str, request: Request):
    input_updates = {}
    try:
        input_updates = await request.json()
        logging.debug("Got input update: ", input_updates)
    except Exception as e:
        logging.debug("Unable to parse request body: ", await request.body(), e)

    payload, _ = await script_task(script_name=user_script, input_updates=input_updates)

    return JSONResponse(payload)


def gen_ws_api_endpoint(user_script):
    async def ws_api_endpoint(websocket):
        await websocket.accept()

        input_updates = {}
        try:
            input_updates = await websocket.receive_json()
            logging.debug("Got input update: ", input_updates)
        except Exception as e:
            logging.debug(
                "Unable to parse request body: ", await websocket.get_text(), e
            )

        payload, _ = await script_task(
            script_name=user_script, input_updates=input_updates
        )
        await websocket.send_json(payload)

        await websocket.close()

    return ws_api_endpoint


async def raster_api_endpoint(user_script: str, request: Request):
    input_updates = {}
    try:
        input_updates = await request.json()
        logging.debug("Got input update: ", input_updates)
    except Exception as e:
        logging.debug("Unable to parse request body: ", await request.body(), e)

    _, payload = await script_task(script_name=user_script, input_updates=input_updates)

    image_bytes_data = payload.read() if payload else None

    return Response(content=image_bytes_data, media_type="image/png")


def get_static_dir_path():
    dist_path = get_python_lib() + "/greppo"
    if os.path.isfile(dist_path):
        BASE_DIR = dist_path
    else:
        BASE_DIR = os.path.dirname(__file__)

    return BASE_DIR + "/static/"


class GreppoServer(object):
    def __init__(self, gr_app: GreppoApp, user_script: str):
        self.gr_app = gr_app
        self.user_script = user_script

    def run(self, host="0.0.0.0", port=8080):
        routes = [
            Route(
                "/api", partial(api_endpoint, self.user_script), methods=["GET", "POST"]
            ),
            Route(
                "/raster",
                partial(raster_api_endpoint, self.user_script),
                methods=["GET", "POST"],
            ),
            WebSocketRoute("/wsapi", gen_ws_api_endpoint(self.user_script)),
            Mount(
                "/",
                app=StaticFiles(directory=get_static_dir_path(), html=True),
                name="static",
            ),
        ]

        middleware = [
            Middleware(
                CORSMiddleware,
                allow_origins=[
                    "http://localhost:8080",
                    "http://0.0.0.0:8080",
                    "http://localhost:8081",
                    "http://0.0.0.0:8081",
                ],
                allow_methods=["GET", "POST"],
            )
        ]

        app = Starlette(debug=True, routes=routes, middleware=middleware)
        uvicorn.run(app, host=host, port=port)

    def close(self):
        pass



from dataclasses import dataclass
from typing import List, Union
import uuid


@dataclass
class WMSTileLayerComponent:
    def __init__(
        self,
        url: str,
        name: str,
        description: str = '',
        visible: bool = True,
        opacity: float = 1.0,
        layers: str = '',
        subdomains: Union[str, List[str]] = '',
        attribution: str = '',
        transparent: bool = True,
        format: str = 'image/jpeg',
    ):
        self.url = url
        self.name = name
        self.description = description
        self.visible = visible
        self.opacity = opacity
        self.layers = layers,
        self.subdomains = subdomains,
        self.attribution = attribution,
        self.transparent = transparent,
        self.format = format,

    def convert_to_dataclass(self):
        id = uuid.uuid4().hex
        return WMSTileLayer(id=id, url=self.url, name=self.name, description=self.description, visible=self.visible, opacity=self.opacity, layers=self.layers, subdomains=self.subdomains, attribution=self.attribution, transparent=self.transparent, format=self.format)


@dataclass()
class WMSTileLayer:
    id: str
    url: str
    name: str
    description: str
    layers: str
    visible: bool
    opacity: float
    subdomains: Union[str, List[str]]
    attribution: str
    transparent: bool
    format: str



import json
from dataclasses import dataclass
from enum import Enum
from typing import Any
from typing import Dict
from typing import List

from geopandas import GeoDataFrame as gdf
from shapely.geometry import LineString
from shapely.geometry import Point
from shapely.geometry import Polygon


default_gdf = gdf(
    [{"id": 9999, "type": "Point", "geometry": Point(0, 0)}], crs="EPSG:4326"
)


@dataclass
class DrawFeature:
    def __init__(
        self,
        name: str,
        geometry: list,
        features: gdf = default_gdf,
        input_updates: Dict[str, Any] = {},
    ):
        self.input_name = name
        self.geometry = geometry
        # TODO Check if it matches EPSG 4326/WSG84
        self.features = features.explode(ignore_index=True)

        self.input_updates = input_updates

    def get_value(self):
        id, name = self.input_name.split("_")
        if name in self.input_updates:
            return draw_feature_dict_2_gdf(self.input_updates.get(name))

        return self.features

    def convert_to_component_info(self):
        _id, name = self.input_name.split("_")
        _type = DrawFeature.__name__
        _geometry = self.geometry
        _features = draw_feature_gdf_2_dict(self.get_value())

        return DrawFeatureComponentInfo(
            id=_id, name=name, type=_type, geometry=_geometry, features=_features
        )

    @classmethod
    def proxy_name(cls):
        return "draw_feature"

    def __repr__(self):
        return str(self.get_value())

    def __str__(self):
        return self.__repr__()


def draw_feature_dict_2_gdf(features_dict):
    features_list = []
    for feature in features_dict:
        _points = []
        _type = feature["type"]
        for latlng in feature["latlngs"]:
            _points.append(Point(latlng["lng"], latlng["lat"]))

        if _type == "LineString":
            feature_geom = LineString(_points)

        if _type == "Polygon":
            feature_geom = Polygon(_points)

        if _type == "Point":
            feature_geom = _points[0]

        features_list.append({"type": _type, "geometry": feature_geom})

    features_gdf = gdf(features_list, crs="EPSG:4326")

    return features_gdf


def draw_feature_gdf_2_dict(features_gdf):
    features_dict = []
    for index, feature in features_gdf.iterrows():
        _type = feature["geometry"].geom_type
        latlngs = []
        if _type == "Polygon":
            for (lng, lat) in list(feature["geometry"].exterior.coords):
                latlngs.append({"lat": lat, "lng": lng})

        if _type == "LineString":
            for (lng, lat) in list(feature["geometry"].coords):
                latlngs.append({"lat": lat, "lng": lng})

        if _type == "Point":
            for (lng, lat) in list(feature["geometry"].coords):
                latlngs.append({"lat": lat, "lng": lng})

        features_dict.append({"type": _type, "latlngs": latlngs})
    return features_dict


@dataclass
class DrawFeatureComponentInfo:
    id: str
    type: str
    name: str
    geometry: list
    features: dict



from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import Union

import numpy as np
import pandas as pd


@dataclass
class Display:
    def __init__(self, name: str, value: str, input_updates: Dict[str, Any] = {}):
        self.input_name = name
        self.value = value

    def get_value(self):
        id, name = self.input_name.split("_")

        return self.value

    def convert_to_component_info(self):
        _id, name = self.input_name.split("_")
        _type = Display.__name__
        value = str(self.get_value())

        return DisplayComponentInfo(id=_id, name=name, type=_type, value=value)

    @classmethod
    def proxy_name(cls):
        return "display"

    def __repr__(self):
        return str(self.get_value())

    def __str__(self):
        return self.__repr__()


@dataclass
class DisplayComponentInfo:
    id: str
    name: str
    type: str
    value: Any



from dataclasses import dataclass
from typing import List

from geopandas import GeoDataFrame as gdf


@dataclass
class ImageLayer:
    id: str
    name: str
    description: str
    url: str
    bounds: List[List]
    visible: bool



from dataclasses import dataclass
from typing import List

from geopandas import GeoDataFrame as gdf


@dataclass
class OverlayLayer:
    id: str
    data: gdf
    name: str
    description: str
    style: dict
    visible: bool
    viewzoom: List



import base64
import copy
import dataclasses
import json
import logging
import uuid
from io import BytesIO
from typing import Any
from typing import Dict
from typing import List, Union

import numpy as np
import rasterio
from geopandas import GeoDataFrame as gdf
from greppo import osm
from PIL import Image
from rasterio.io import MemoryFile
from rasterio.warp import calculate_default_transform
from rasterio.warp import reproject

from .input_types import BarChart
from .input_types import ComponentInfo
from .input_types import DrawFeature
from .input_types import GreppoInputs
from .input_types import LineChart
from .input_types import Multiselect
from .input_types import Number
from .input_types import Select
from .input_types import Text
from .input_types import Display

from .layers.base_layer import BaseLayerComponent, BaseLayer
from .layers.tile_layer import TileLayer, TileLayerComponent
from .layers.wms_tile_layer import WMSTileLayer, WMSTileLayerComponent
from .layers.vector_layer import VectorLayer, VectorLayerComponent
from .layers.image_layer import ImageLayer
from .layers.overlay_layer import OverlayLayer
from .layers.ee_layer import EarthEngineLayerComponent


class GreppoApp(object):
    """
    The main Greppo class that is the entry point for user scripts. User scripts will use this class via a module
    import variable `gpo`.

    This class provides an interface around available frontend component elements. The methods simply point to the
    backend representation of those frontend component elements (ie. `Number` is the backend class that a user script
    can access via `self.number`.
    """

    def __init__(self, name: str = "Untitled App"):
        self.name: str = name
        self.display = Display
        self.select = Select
        self.multiselect = Multiselect
        self.draw_feature = DrawFeature
        self.bar_chart = BarChart
        self.line_chart = LineChart

    # UX component proxy methods
    @staticmethod
    def number():
        """
        Interactive Number value rendered on the frontend.
        """
        return Number

    @staticmethod
    def text():
        """
        Interactive Text value rendered on the frontend.
        """
        return Text


class GreppoAppProxy(object):
    """
    Proxy object that mirrors the `GreppoApp` class. Adds additional methods that user scripts don't need to know about.
    These methods are used by a Greppo server to obtain an output from the user script that is then rendered by the
    frontend.
    """

    def __init__(self):
        # Map component data
        self.map_data: Dict = {'settings': {'zoom': 3, 'center': [0, 0], 'maxZoom': 18, 'minZoom': 0}}
        self.base_layers: List[BaseLayer] = []
        self.tile_layers: List[TileLayer] = []
        self.wms_tile_layers: List[WMSTileLayer] = []
        self.vector_layers: List[VectorLayer] = []
        self.image_layers: List[ImageLayer] = []
        # TODO Cleanup raster temp
        self.raster_image_reference: List[bytes] = []
        self.registered_inputs: List[ComponentInfo] = []

        # Input updates
        self.inputs = {}

    def display(self, **kwargs):
        display = Display(**kwargs)
        self.register_input(display)
        return display

    def number(self, **kwargs):
        number = Number(**kwargs)
        self.register_input(number)
        return number

    def text(self, **kwargs):
        text = Text(**kwargs)
        self.register_input(text)
        return text

    def select(self, **kwargs):
        select = Select(**kwargs)
        self.register_input(select)
        return select

    def multiselect(self, **kwargs):
        multiselect = Multiselect(**kwargs)
        self.register_input(multiselect)
        return multiselect

    def draw_feature(self, **kwargs):
        draw_feature = DrawFeature(**kwargs)
        self.register_input(draw_feature)
        return draw_feature

    def bar_chart(self, **kwargs):
        bar_chart = BarChart(**kwargs)
        self.register_input(bar_chart)
        return bar_chart

    def line_chart(self, **kwargs):
        line_chart = LineChart(**kwargs)
        self.register_input(line_chart)
        return line_chart

    def map(self, **kwargs):        
        if 'zoom' in kwargs:
            self.map_data['settings']['zoom'] = kwargs.get('zoom')
        if 'center' in kwargs:
            self.map_data['settings']['center'] = kwargs.get('center')
        if 'max_zoom' in kwargs:
            self.map_data['settings']['maxZoom'] = kwargs.get('max_zoom')
        if 'min_zoom' in kwargs:
            self.map_data['settings']['minZoom'] = kwargs.get('min_zoom')

    def ee_layer(self, **kwargs):
        ee_layer_component = EarthEngineLayerComponent(**kwargs)
        ee_layer_dataclass = ee_layer_component.convert_to_dataclass()
        self.tile_layers.append(ee_layer_dataclass)

    def tile_layer(self, **kwargs):
        tile_layer_component = TileLayerComponent(**kwargs)
        tile_layer_dataclass = tile_layer_component.convert_to_dataclass()
        self.tile_layers.append(tile_layer_dataclass)

    def wms_tile_layer(self, **kwargs):
        wms_tile_layer_component = WMSTileLayerComponent(**kwargs)
        wms_tile_layer_dataclass = wms_tile_layer_component.convert_to_dataclass()
        self.wms_tile_layers.append(wms_tile_layer_dataclass)

    def base_layer(
        self,
        **kwargs
    ):
        base_layer_component = BaseLayerComponent(**kwargs)
        base_layer_dataclass = base_layer_component.convert_to_dataclass()
        self.base_layers.append(base_layer_dataclass)

    def vector_layer(self, **kwargs):
        vector_layer_component = VectorLayerComponent(**kwargs)
        vector_layer_dataclass = vector_layer_component.convert_to_dataclass()
        self.vector_layers.append(vector_layer_dataclass)

    def vector_layer(self, **kwargs):
        vector_layer_component = VectorLayerComponent(**kwargs)
        vector_layer_dataclass = vector_layer_component.convert_to_dataclass()
        self.vector_layers.append(vector_layer_dataclass)

    def overlay_layer(
        self, data: gdf, name: str, description: str, style: dict, visible: bool
    ):
        vector_layer_component = VectorLayerComponent(
            data=data, name=name, description=description, style=style, visible=visible)
        vector_layer_dataclass = vector_layer_component.convert_to_dataclass()
        self.vector_layers.append(vector_layer_dataclass)

    def raster_layer(self, file_path: str, name: str, description: str, visible: bool):
        id = uuid.uuid4().hex

        src_dataset = rasterio.open(file_path)
        dst_crs = "EPSG:4326"

        transform, width, height = calculate_default_transform(
            src_dataset.crs,
            dst_crs,
            src_dataset.width,
            src_dataset.height,
            *src_dataset.bounds
        )

        dst_bands = []
        for band_n_1 in range(src_dataset.count):
            src_band = rasterio.band(src_dataset, band_n_1 + 1)
            dst_band = reproject(src_band, dst_crs=dst_crs)
            dst_bands.append(dst_band)

        if src_dataset.count != 3:
            for i in range(len(dst_bands), src_dataset.count):
                dst_bands.append(rasterio.band(src_dataset, 1))

        alpha = np.where(dst_bands[0][0] > 1e8, 0, 1)
        alpha_band = list(copy.deepcopy(dst_bands[0]))
        alpha_band[0] = alpha.astype("uint8")
        dst_bands.append(tuple(alpha_band))

        png_kwargs = src_dataset.meta.copy()
        png_kwargs.update(
            {
                "crs": dst_crs,
                "width": width,
                "height": height,
                "driver": "PNG",
                "dtype": rasterio.uint8,
                "transform": transform,
                "count": len(dst_bands),
            }
        )

        with MemoryFile() as png_memfile:
            with png_memfile.open(**png_kwargs) as dst_file:
                for i_1, dst_band in enumerate(dst_bands):
                    dst_file.write(dst_band[0][0], i_1 + 1)

                    dst_file.write_colormap(
                        i_1 + 1, {0: (255, 0, 0, 255), 255: (0, 0, 0, 255)}
                    )

                self.raster_image_reference.append(png_memfile.read())

            url = (
                "data:image/png;base64," +
                base64.b64encode(png_memfile.read()).decode()
            )
            (bounds_bottom, bounds_right) = transform * (0, 0)
            (bounds_top, bounds_left) = transform * (width, height)
            bounds = [[bounds_left, bounds_bottom], [bounds_right, bounds_top]]

            self.image_layers.append(
                ImageLayer(id, name, description, url, bounds, visible)
            )

    def image_layer(self, file_path: str, name: str, description: str, visible: bool):
        id = uuid.uuid4().hex

        file_ext = file_path.split(".")[-1].lower()
        assert file_ext in [
            "png",
            "jpg",
            "jpeg",
        ], "Image input extension should be png, jpg or jpeg for image_layer"

        buffered = BytesIO()
        image = Image.open(file_path)
        image.save(buffered, format="JPEG")
        image_string = base64.b64encode(buffered.getvalue()).decode()

        url = "data:image/png;base64," + image_string

        bounds = [[0, 0], [100, 100]]
        bounds = [
            [14.760840184106792, 77.97900023926854],
            [14.763995704693206, 77.98389492733145],
        ]

        self.image_layers.append(
            ImageLayer(id, name, description, url, bounds, visible)
        )

    def update_inputs(self, inputs: Dict[str, Any]):
        self.inputs = inputs

    def register_input(self, discovered_input: GreppoInputs):
        """
        BarChart and LineChart are also registered with this `register_input` method. Maybe rename this method.
        """
        component_info = discovered_input.convert_to_component_info()
        self.registered_inputs.append(component_info)

        return discovered_input

    def gpo_prepare_data(self):
        """
        Take output of run script and setup the payload for the front-end to read.
        """

        app_output = {
            "base_layer_data": [],
            "tile_layer_data": [],
            "wms_tile_layer_data": [],
            "vector_layer_data": [],
            "image_layer_data": [],
            "component_info": [],
            "map": {},
        }
        app_output["map"] = self.map_data
        
        for _tile_layer in self.tile_layers:
            s = {}
            for k, v in _tile_layer.__dict__.items():
                _v = v
                s[k] = _v
            app_output["tile_layer_data"].append(s)

        for _wms_tile_layer in self.wms_tile_layers:
            s = {}
            for k, v in _wms_tile_layer.__dict__.items():
                _v = v
                s[k] = _v
            app_output["wms_tile_layer_data"].append(s)

        for _base_layer in self.base_layers:
            s = {}
            for k, v in _base_layer.__dict__.items():
                _v = v
                s[k] = _v
            app_output["base_layer_data"].append(s)

        for _vector_layer in self.vector_layers:
            s = {}
            for k, v in _vector_layer.__dict__.items():
                _v = v
                if k == "data":
                    _v = json.loads(v.to_json())
                s[k] = _v

            app_output["vector_layer_data"].append(s)

        for _image_layer in self.image_layers:
            s = {}
            for k, v in _image_layer.__dict__.items():
                _v = v
                s[k] = _v
            app_output["image_layer_data"].append(s)

        app_output["component_info"] = [
            dataclasses.asdict(i) for i in self.registered_inputs
        ]

        logging.info("Len component info: ", len(app_output["component_info"]))

        return app_output

    def gpo_reference_data(self):
        """ Only return one reference image for testing. """
        if len(self.raster_image_reference) == 0:
            return None
        return self.raster_image_reference[0]


app = GreppoApp()



from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import Union

import numpy as np
import pandas as pd

@dataclass
class Text:
    def __init__(self, name: str, value: str, input_updates: Dict[str, Any] = {}):
        self.input_name = name
        self.value = value

        self.input_updates = input_updates

    def get_value(self):
        id, name = self.input_name.split("_")
        return self.input_updates.get(name, self.value)

    def convert_to_component_info(self):
        _id, name = self.input_name.split("_")
        _type = Text.__name__
        value = str(self.get_value())

        return TextComponentInfo(id=_id, name=name, type=_type, value=value)

    @classmethod
    def proxy_name(cls):
        return "text"

    def __add__(self, other):
        if type(other) in [str]:
            return self.get_value() + other
        elif type(other) is Text:
            return self.get_value() + other.get_value()
        else:
            raise Exception("Operation Not Supported for type: " + str(type(other)))

    __radd__ = __add__

    def __mul__(self, other):
        if type(other) in [str]:
            return self.get_value() * other
        elif type(other) in [list, str, np.array, np.ndarray, pd.Series]:
            return other * self.get_value()
        elif type(other) is Text:
            return self.get_value() * other.get_value()
        else:
            raise Exception("Operation Not Supported for type: " + str(type(other)))

    __rmul__ = __mul__

    def __repr__(self):
        return str(self.get_value())

    def __str__(self):
        return self.__repr__()


@dataclass
class TextComponentInfo:
    id: str
    name: str
    type: str
    value: Any



import ast
import io
import json
import logging
import secrets
from _ast import Assign
from _ast import Attribute
from _ast import Call
from _ast import keyword
from _ast import Load
from _ast import Name
from _ast import Store
from ast import Str
from contextlib import redirect_stdout
from typing import Any, Dict

from greppo import GreppoAppProxy
from .input_types import GreppoInputsNames, GreppoChartNames
from meta.asttools import print_ast, dump_python_source

logger = logging.getLogger('user_script_utils')


class RenameGreppoAppTransformer(ast.NodeTransformer):
    def __init__(self, hash_prefix):
        super().__init__()

        self.hash_prefix = hash_prefix

    def visit_Name(self, node):
        if node.id == "app":
            node.id = self.hash_prefix + "_app"

        return node


def append_send_data_method(code):
    code.body.append(
        Assign(
            targets=[Name(ctx=Store(), id="gpo_payload")],
            type_comment=None,
            value=Call(
                args=[],
                func=Attribute(
                    attr=GreppoAppProxy.gpo_prepare_data.__name__,
                    ctx=Load(),
                    value=Name(ctx=Load(), id="app"),
                ),
                keywords=[],
            ),
        )
    )

    return code


def append_raster_reference(code):
    code.body.append(
        Assign(
            targets=[Name(ctx=Store(), id="gpo_raster_reference_payload")],
            type_comment=None,
            value=Call(
                args=[],
                func=Attribute(
                    attr=GreppoAppProxy.gpo_reference_data.__name__,
                    ctx=Load(),
                    value=Name(ctx=Load(), id="app"),
                ),
                keywords=[],
            ),
        )
    )

    return code


class AddInputUpdatesToGpoVariableAndGetValueTransformer(ast.NodeTransformer):
    def __init__(self, input_updates: Dict[str, Any], hex_token_generator):
        super().__init__()

        self.input_updates = input_updates

        self.hex_token_generator = hex_token_generator

        self.greppo_input_calls = []

    def visit_Call(self, node):
        if not hasattr(node.func, "attr"):
            return node

        if node.func.attr not in GreppoInputsNames:
            return node

        # ==== Find name value for gpo variable and set name prefix
        for node_kwargs in node.keywords:
            if node_kwargs.arg == "name":
                string_container = node_kwargs.value  ## 3.6 uses an object called Str that has `s` instead of `value`.
                if type(string_container) == Str:
                    name = string_container.s
                else:
                    name = string_container.value

                input_name = self.hex_token_generator(nbytes=4) + "_" + name

                if type(string_container) == Str:
                    node_kwargs.value.s = input_name
                else:
                    node_kwargs.value.value = input_name

                break

        # ==== Add input_updates to the ctr.
        input_updates_kwarg = keyword(
            arg='input_updates',
            value=ast.parse(json.dumps(self.input_updates)).body[0].value
        )
        node.keywords.append(input_updates_kwarg)

        # ==== Create new Call node for get_value
        z = Call(args=[],
                    func=Attribute(
                        attr='get_value',
                        ctx=Load(),
                        value=node
                    ),
                    keywords=[])

        return z


class AddHexPrefixForCharts(ast.NodeTransformer):
    def __init__(self, input_updates, hex_token_generator):
        super().__init__()

        self.input_updates = input_updates

        self.hex_token_generator = hex_token_generator

    def visit_Call(self, node):
        if not hasattr(node.func, "attr"):
            return node

        if node.func.attr not in GreppoChartNames:
            return node

        # ==== Find all names for gpo_inputs and set hex id
        for node_kwargs in node.keywords:
            if node_kwargs.arg == "name":
                string_container = node_kwargs.value  ## 3.6 uses an object called Str that has `s` instead of `value`.
                if type(string_container) == Str:
                    input_name = string_container.s
                else:
                    input_name = string_container.value

                input_name = self.hex_token_generator(nbytes=4) + "_" + input_name

                if type(string_container) == Str:
                    node_kwargs.value.s = input_name
                else:
                    node_kwargs.value.value = input_name

                break

        input_updates_ast = ast.parse(str(self.input_updates)).body[0]
        if not hasattr(input_updates_ast, 'value'):
            raise Exception("Cannot parse input_updates: {}", self.input_updates)

        input_updates_keyword = keyword(
            arg="input_updates", value=input_updates_ast.value
        )

        # ==== Add input updates to node
        updated = False
        for pos, k in enumerate(node.keywords):
            if k.arg == "input_updates":
                node.keywords[pos] = input_updates_keyword
                updated = True
                break
        if not updated:
            node.keywords.append(input_updates_keyword)

        return node


def run_script(script_name, input_updates, hex_token_generator):
    with open(script_name) as f:
        lines = f.read()
        user_code = ast.parse(lines, script_name)

        add_input_updates_to_gpo_variable_and_get_value_t = AddInputUpdatesToGpoVariableAndGetValueTransformer(
            input_updates=input_updates, hex_token_generator=hex_token_generator
        )
        add_input_updates_to_gpo_variable_and_get_value_t.visit(user_code)
        ast.fix_missing_locations(
            user_code
        )

        transformer = AddHexPrefixForCharts(
            input_updates=input_updates, hex_token_generator=hex_token_generator
        )
        transformer.visit(user_code)
        ast.fix_missing_locations(
            user_code
        )

        user_code = append_send_data_method(user_code)
        user_code = append_raster_reference(user_code)

        # Transform gpo for locals() injection
        hash_prefix = hex_token_generator(nbytes=4)
        gpo_transformer = RenameGreppoAppTransformer(hash_prefix=hash_prefix)
        gpo_transformer.visit(user_code)

        ast.fix_missing_locations(user_code)

        #logger.debug('\n\n------ Code Transform ------\n')
        #logger.debug(dump_python_source(user_code))
        #logger.debug('\n----------------------------\n\n')

        gpo_app = GreppoAppProxy()

        locals_copy = locals().copy()
        locals_copy['app'] = gpo_app

        # TODO maybe have a fresh locals obj?
        locals()[hash_prefix + "_app"] = gpo_app

        exec(compile(user_code, script_name, "exec"), globals(), locals())

        raster_reference_payload = locals().get("gpo_raster_reference_payload", None)

        return locals()["gpo_payload"], raster_reference_payload


async def script_task(
    script_name: str,
    input_updates: Dict[str, Any],
    hex_token_generator=secrets.token_hex,
):
    """
    async task that runs a user_script in a Greppo context (`gpo_send_data`) and generates payload for front-end
    consumption.
    """
    logging.info("Loading Greppo App at: " + script_name)

    with redirect_stdout(io.StringIO()) as loop_out:
        payload = run_script(
            script_name=script_name,
            input_updates=input_updates,
            hex_token_generator=hex_token_generator,
        )

    logger.setLevel(logging.DEBUG)

    # logger.info("-------------")
    # logger.info("stdout from process")
    # logger.info("===")
    # logger.info(loop_out.getvalue())
    # logger.info("===")

    return payload
